from dateutil.easter import EASTER_WESTERN

from .holidays import Country


class BE(Country):
    id = "BE"
    languages = ["de", "fr", "nl"]
    easter_type = EASTER_WESTERN

    def __init__(self):
        super().__init__()

        self.define_holiday() \
            .with_names({
                "de": "Neujahr",
                "fr": "Nouvel An",
                "nl": "Nieuwjaar",
            }) \
            .on("01-01") \
            .with_flags("NF")

        self.define_holiday() \
            .with_names({
                "de": "Tag der Arbeit",
                "fr": "Fête du Travail",
                "nl": "Dag van de arbeid",
            }) \
            .on("05-01") \
            .with_flags("NF")

        self.define_holiday() \
            .with_names({
                "de": "Nationalfeiertag",
                "fr": "Fête nationale",
                "nl": "Nationale feestdag",
            }) \
            .on("07-21") \
            .with_flags("NF")

        self.define_holiday() \
            .with_names({
                "de": "Mariä Himmelfahrt",
                "fr": "Assomption",
                "nl": "Onze Lieve Vrouw hemelvaart",
            }) \
            .on("08-15") \
            .with_flags("NRF")

        self.define_holiday() \
            .with_names({
                "de": "Allerheiligen",
                "fr": "Toussaint",
                "nl": "Allerheiligen",
            }) \
            .on("11-01") \
            .with_flags("NRF")

        self.define_holiday() \
            .with_names({
                "de": "Waffenstillstand",
                "fr": "Jour de l'armistice",
                "nl": "Wapenstilstand",
            }) \
            .on("11-11") \
            .with_flags("NF")

        self.define_holiday() \
            .with_names({
                "de": "Weihnacht",
                "fr": "Noël",
                "nl": "Kerstmis",
            }) \
            .on("12-25") \
            .with_flags("NRF")

        self.define_holiday() \
            .with_names({
                "de": "Ostern",
                "fr": "Pâques",
                "nl": "Pasen",
            }) \
            .on("Easter") \
            .with_flags("NRV")

        self.define_holiday() \
            .with_names({
                "de": "Ostermontag",
                "fr": "Lundi de Pâques",
                "nl": "Paasmaandag",
            }) \
            .on("1 day after Easter") \
            .with_flags("NRV")

        self.define_holiday() \
            .with_names({
                "de": "Christi Himmelfahrt",
                "fr": "Ascension",
                "nl": "Onze Lieve Heer hemelvaart",
            }) \
            .on("39 days after Easter") \
            .with_flags("NRV")

        self.define_holiday() \
            .with_names({
                "de": "Pfingsten",
                "fr": "Pentecôte",
                "nl": "Pinksteren",
            }) \
            .on("49 days after Easter") \
            .with_flags("NRV")

        self.define_holiday() \
            .with_names({
                "de": "Pfingstmontag",
                "fr": "Lundi de Pentecôte",
                "nl": "Pinkstermaandag",
            }) \
            .on("50 days after Easter") \
            .with_flags("NRV")



from holidata.holidays.holidays import Region


class VC(Region):
    def __init__(self, country):
        super().__init__("VC", country)

        self.define_holiday() \
            .with_name("Lunes de Fallas") \
            .in_years([2013]) \
            .on("03-18") \
            .with_flags("F")

        self.define_holiday() \
            .with_name("San Juan") \
            .in_years([2019, 2020, 2021, 2022, 2023, 2024]) \
            .on("06-24") \
            .with_flags("RF")

        self.define_holiday() \
            .with_name("Día de la Comunitat Valenciana") \
            .except_for([2011, 2016, 2022]) \
            .on("10-09") \
            .with_flags("F")

        self.define_holiday() \
            .with_name("San José") \
            .in_years([2011, 2012, 2013, 2014, 2015, 2016, 2018, 2019, 2020, 2021, 2022, 2024]) \
            .on("03-19") \
            .with_flags("RF")

        self.define_holiday() \
            .with_name("Jueves Santo") \
            .in_years([2011, 2016, 2017, 2022, 2023]) \
            .on("3 days before Easter") \
            .with_flags("RV")

        self.define_holiday() \
            .with_name("Lunes de Pascua") \
            .except_for([2023]) \
            .on("1 day after Easter") \
            .with_flags("RV")

        self.define_holiday() \
            .with_name("Lunes siguiente a la Fiesta del Trabajo") \
            .in_years([2011]) \
            .on("05-02") \
            .with_flags("F")

        self.define_holiday() \
            .with_name("Lunes siguiente al Día de la Constitución Española") \
            .in_years([2015]) \
            .on("12-07") \
            .with_flags("F")

        self.define_holiday() \
            .with_name("San Esteban") \
            .in_years([2016]) \
            .on("12-26") \
            .with_flags("RF")



from dateutil.easter import EASTER_WESTERN

from holidata.utils import month_reference
from .holidays import Country


class US(Country):
    id = "US"
    languages = ["en", "es"]
    default_lang = "en"
    regions = ["AL", "AK", "AZ", "AR", "CA", "CO", "CT", "DE", "FL", "GA", "HI", "ID", "IL", "IN", "IA", "KS", "KY",
               "LA", "ME", "MD", "MA", "MI", "MN", "MS", "MO", "MT", "NE", "NV", "NH", "NJ", "NM", "NY", "NC", "ND",
               "OH", "OK", "OR", "PA", "RI", "SC", "SD", "TN", "TX", "UT", "VT", "VA", "WA", "WV", "WI", "WY", "DC",
               "AS", "GU", "MP", "PR", "UM", "VI"]
    easter_type = EASTER_WESTERN

    def __init__(self):
        super().__init__()

        self.define_holiday() \
            .with_names({
                "en": "New Year's Day",
                "es": "Año Neuvo",
            }) \
            .on("01-01") \
            .with_flags("NF")

        self.define_holiday() \
            .with_names({
                "en": "Independence Day",
                "es": "Día de la Independiencia",
            }) \
            .on("07-04") \
            .with_flags("NF")

        self.define_holiday() \
            .with_names({
                "en": "Veterans Day",
                "es": "Día de los Veteranos",
            }) \
            .on("11-11") \
            .with_flags("NF")

        self.define_holiday() \
            .with_names({
                "en": "Christmas Eve",
                "es": "Nochebuena",
            }) \
            .on("12-24") \
            .with_flags("NRF")

        self.define_holiday() \
            .with_names({
                "en": "Christmas Day",
                "es": "Navidad",
            }) \
            .on("12-25") \
            .with_flags("NRF")

        self.define_holiday() \
            .with_names({
                "en": "Birthday of Martin Luther King, Jr.",
                "es": "Cumpleaños de Martin Luther King, Jr.",
            }) \
            .on("3. monday in January") \
            .with_flags("NV")

        self.define_holiday() \
            .with_names({
                "en": "Washington's Birthday",
                "es": "Día del Presidente",
            }) \
            .on("3. monday in February") \
            .with_flags("NV")

        self.define_holiday() \
            .with_names({
                "en": "Patriots' Day",
                "es": "Día del Patriota",
                }) \
            .in_regions(["MA", "ME"]) \
            .on("3. monday in April") \
            .with_flags("V")

        self.define_holiday() \
            .with_names({
                "en": "Memorial Day",
                "es": "Día de los Caídos",
            }) \
            .on("1. last monday in May") \
            .with_flags("NV")

        self.define_holiday() \
            .with_names({
                "en": "Labor Day",
                "es": "Día del Trabajo",
            }) \
            .on("1. monday in September") \
            .with_flags("NV")

        self.define_holiday() \
            .with_names({
                "en": "Columbus Day",
                "es": "Día de Columbus",
            }) \
            .on("2. monday in October") \
            .with_flags("NV")

        self.define_holiday() \
            .with_names({
                "en": "Thanksgiving Day",
                "es": "Día de Acción de Gracias",
            }) \
            .on("4. thursday in November") \
            .with_flags("NV")

        self.define_holiday() \
            .with_names({
                "en": "Day after Thanksgiving",
                "es": "Día después de Acción de Gracias",
            }) \
            .on(self.day_after_thanksgiving) \
            .with_flags("NV")

    @staticmethod
    def day_after_thanksgiving(year):
        date = month_reference(year, "november").shift_to_weekday(
            "thursday",
            order=4,
            reverse=False,
            including=True,
        )

        return date.shift(days=1)




from holidata.holidays.holidays import Region


class EX(Region):
    def __init__(self, country):
        super().__init__("EX", country)

        self.define_holiday() \
            .with_name("Lunes siguiente a la Epifanía del Señor") \
            .in_years([2013, 2019]) \
            .on("01-07") \
            .with_flags("RF")

        self.define_holiday() \
            .with_name("Martes de Carnaval") \
            .in_years([2023, 2024]) \
            .on("47 days before Easter") \
            .with_flags("V")

        self.define_holiday() \
            .with_name("Lunes siguiente a San José") \
            .in_years([2017]) \
            .on("03-20") \
            .with_flags("RF")

        self.define_holiday() \
            .with_name("Día de Extremadura") \
            .except_for([2019, 2024]) \
            .on("09-08") \
            .with_flags("F")

        self.define_holiday() \
            .with_name("Lunes siguiente al Día de Extremadura") \
            .in_years([2013, 2019]) \
            .on("09-09") \
            .with_flags("F")

        self.define_holiday() \
            .with_name("San José") \
            .in_years([2021]) \
            .on("03-19") \
            .with_flags("RF")

        self.define_holiday() \
            .with_name("Lunes siguiente a la Fiesta Nacional de España") \
            .in_years([2014]) \
            .on("10-13") \
            .with_flags("F")

        self.define_holiday() \
            .with_name("Lunes siguiente a Todos los Santos") \
            .in_years([2015, 2020]) \
            .on("11-02") \
            .with_flags("F")

        self.define_holiday() \
            .with_name("Lunes siguiente al Día de la Constitución Española") \
            .in_years([2015, 2020]) \
            .on("12-07") \
            .with_flags("F")

        self.define_holiday() \
            .with_name("Lunes siguiente a La Inmaculada Concepción") \
            .in_years([2013, 2019, 2024]) \
            .on("12-09") \
            .with_flags("RF")

        self.define_holiday() \
            .with_name("Jueves Santo") \
            .on("3 days before Easter") \
            .with_flags("RV")

        self.define_holiday() \
            .with_name("Lunes siguiente a la Fiesta del Trabajo") \
            .in_years([2011, 2016, 2022]) \
            .on("05-02") \
            .with_flags("F")

        self.define_holiday() \
            .with_name("Lunes siguiente a la Natividad del Señor") \
            .in_years([2022]) \
            .on("12-26") \
            .with_flags("RF")

        self.define_holiday() \
            .with_name("San Esteban") \
            .in_years([2011, 2016]) \
            .on("12-26") \
            .with_flags("RF")



from holidata.holidays.holidays import Region


class CB(Region):
    def __init__(self, country):
        super().__init__("CB", country)

        self.define_holiday() \
            .with_name("Lunes siguiente a la Epifanía del Señor") \
            .in_years([2013]) \
            .on("01-07") \
            .with_flags("RF")

        self.define_holiday() \
            .with_name("Día de las Instituciones de Cantabria") \
            .in_years([2011, 2016, 2017, 2018, 2020, 2021, 2022, 2023]) \
            .on("07-28") \
            .with_flags("F")

        self.define_holiday() \
            .with_name("La Bien Aparecida") \
            .in_years([2011, 2012, 2014, 2014, 2015, 2016, 2017, 2018, 2020, 2021, 2022, 2023]) \
            .on("09-15") \
            .with_flags("RF")

        self.define_holiday() \
            .with_name("Jueves Santo") \
            .in_years([2011, 2012, 2013, 2014, 2015, 2016, 2017, 2019, 2020, 2021, 2022, 2023, 2024]) \
            .on("3 days before Easter") \
            .with_flags("RV")

        self.define_holiday() \
            .with_name("Lunes de Pascua") \
            .in_years([2013, 2015, 2019, 2020, 2024]) \
            .on("1 day after Easter") \
            .with_flags("RV")

        self.define_holiday() \
            .with_name("Lunes siguiente a la Fiesta del Trabajo") \
            .in_years([2011]) \
            .on("05-02") \
            .with_flags("F")

        self.define_holiday() \
            .with_name("Santiago Apóstol") \
            .in_years([2012, 2013, 2014, 2019]) \
            .on("07-25") \
            .with_flags("RF")

        self.define_holiday() \
            .with_name("Santiago Apóstol / Día Nacional de Galicia") \
            .in_years([2024]) \
            .on("07-25") \
            .with_flags("RF")

        self.define_holiday() \
            .with_name("Lunes siguiente a Todos los Santos") \
            .in_years([2015]) \
            .on("11-02") \
            .with_flags("F")

        self.define_holiday() \
            .with_name("Lunes siguiente a La Inmaculada Concepción") \
            .in_years([2019]) \
            .on("12-09") \
            .with_flags("RF")

        self.define_holiday() \
            .with_name("Lunes siguiente a la Natividad del Señor") \
            .in_years([2022]) \
            .on("12-26") \
            .with_flags("RF")

        self.define_holiday() \
            .with_name("San Esteban") \
            .in_years([2016]) \
            .on("12-26") \
            .with_flags("RF")



from dateutil.easter import EASTER_ORTHODOX

from .holidays import Country
from holidata.utils import SmartDayArrow, easter


class GR(Country):
    id = "GR"
    languages = ["el"]
    default_lang = "el"
    easter_type = EASTER_ORTHODOX

    def __init__(self):
        super().__init__()

        self.define_holiday() \
            .with_name("Πρωτοχρονιά") \
            .on("01-01") \
            .with_flags("NF")

        self.define_holiday() \
            .with_name("Θεοφάνεια") \
            .on("01-06") \
            .with_flags("NRF")

        self.define_holiday() \
            .with_name("Ευαγγελισμός της Θεοτόκου και Εθνική Ημέρα Ανεξαρτησίας της Ελλάδας") \
            .on("03-25") \
            .with_flags("NF")

        self.define_holiday() \
            .with_name("Κοίμηση της Θεοτόκου") \
            .on("08-15") \
            .with_flags("NRF")

        self.define_holiday() \
            .with_name("Ημέρα του ΌΧΙ") \
            .on("10-28") \
            .with_flags("NF")

        self.define_holiday() \
            .with_name("Χριστούγεννα") \
            .on("12-25") \
            .with_flags("NRF")

        self.define_holiday() \
            .with_name("Επόμενη ημέρα Χριστουγέννων") \
            .on("12-26") \
            .with_flags("NRF")

        self.define_holiday() \
            .with_name("Καθαρά Δευτέρα") \
            .on("48 days before Easter") \
            .with_flags("NRV")

        self.define_holiday() \
            .with_name("Μεγάλη Παρασκευή") \
            .on("2 days before Easter") \
            .with_flags("NRV")

        self.define_holiday() \
            .with_name("Μεγάλο Σάββατο") \
            .on("1 day before Easter") \
            .with_flags("NRV")

        self.define_holiday() \
            .with_name("Πάσχα") \
            .on("Easter") \
            .with_flags("NRV")

        self.define_holiday() \
            .with_name("Δευτέρα του Πάσχα") \
            .on("1 day after Easter") \
            .with_flags("NRV")

        self.define_holiday() \
            .with_name("Δευτέρα του Αγίου Πνεύματος") \
            .on("50 days after Easter") \
            .with_flags("NRV")

        self.define_holiday() \
            .with_name("Πρωτομαγιά") \
            .on(self.date_for_may_day) \
            .with_flags("NF")

    def date_for_may_day(self, year):
        """
        Postponed if it collides with Easter
        """
        date = SmartDayArrow(year, 5, 1)
        easter_date = easter(year, self.easter_type)

        if date == easter_date:
            date = date.shift(days=2)
        elif date == easter_date.shift(days=1):
            date = date.shift(days=1)

        return date



from dateutil.easter import EASTER_WESTERN

from .holidays import Country

"""
sources
https://www.riigiteataja.ee/akt/109032011007 (Public Holidays and Days of National Importance Act)
"""


class EE(Country):
    id = "EE"
    languages = ["et"]
    default_lang = "et"
    easter_type = EASTER_WESTERN

    def __init__(self):
        super().__init__()

        self.define_holiday() \
            .with_name("Uusaasta") \
            .on("01-01") \
            .with_flags("NF")

        self.define_holiday() \
            .with_name("Iseseisvuspäev, Eesti Vabariigi aastapäev") \
            .on("02-24") \
            .with_flags("NF")

        self.define_holiday() \
            .with_name("Kevadpüha") \
            .on("05-01") \
            .with_flags("NF")

        self.define_holiday() \
            .with_name("Võidupüha") \
            .on("06-23") \
            .with_flags("NF")

        self.define_holiday() \
            .with_name("Jaanipäev") \
            .on("06-24") \
            .with_flags("NF")

        self.define_holiday() \
            .with_name("Taasiseseisvumispäev") \
            .on("08-20") \
            .with_flags("NF")

        self.define_holiday() \
            .with_name("Jõululaupäev") \
            .on("12-24") \
            .with_flags("NF")

        self.define_holiday() \
            .with_name("Esimene jõulupüha") \
            .on("12-25") \
            .with_flags("NF")

        self.define_holiday() \
            .with_name("Teine jõulupüha") \
            .on("12-26") \
            .with_flags("NF")

        self.define_holiday() \
            .with_name("Suur reede") \
            .on("2 days before Easter") \
            .with_flags("NRV")

        self.define_holiday() \
            .with_name("Ülestõusmispühade 1. püha") \
            .on("Easter") \
            .with_flags("NRV")

        self.define_holiday() \
            .with_name("Nelipühade 1. püha") \
            .on("49 days after Easter") \
            .with_flags("NRV")



from dateutil.easter import EASTER_WESTERN

from holidata.utils import SmartDayArrow
from .holidays import Country


class DE(Country):
    id = "DE"
    languages = ["de"]
    default_lang = "de"
    regions = ["BB", "BE", "BW", "BY", "HB", "HE", "HH", "MV", "NI", "NW", "RP", "SH", "SL", "SN", "ST", "TH"]
    easter_type = EASTER_WESTERN

    def __init__(self):
        super().__init__()
        self.default_lang = "de"

        self.define_holiday() \
            .with_name("Neujahr") \
            .on("01-01") \
            .with_flags("NF")

        self.define_holiday() \
            .with_name("Heilige drei Könige") \
            .in_regions(["BW", "BY", "ST"]) \
            .on("01-06") \
            .with_flags("RF")

        self.define_holiday() \
            .with_name("Erster Maifeiertag") \
            .on("05-01") \
            .with_flags("NF")

        self.define_holiday() \
            .with_name("Mariä Himmelfahrt") \
            .in_region("SL") \
            .on("08-15") \
            .with_flags("RF")

        self.define_holiday() \
            .with_name("Tag der Deutschen Einheit") \
            .on("10-03") \
            .with_flags("NRF")

        self.define_holiday() \
            .with_name("Allerheiligen") \
            .in_regions(["BW", "BY", "NW", "RP", "SL"]) \
            .on("11-01") \
            .with_flags("RF")

        self.define_holiday() \
            .with_name("Heilig Abend") \
            .on("12-24") \
            .with_flags("NRF")

        self.define_holiday() \
            .with_name("Weihnachtstag") \
            .on("12-25") \
            .with_flags("NRF")

        self.define_holiday() \
            .with_name("Zweiter Weihnachtstag") \
            .on("12-26") \
            .with_flags("NRF")

        self.define_holiday() \
            .with_name("Silvester") \
            .on("12-31") \
            .with_flags("NF")

        self.define_holiday() \
            .with_name("Karfreitag") \
            .on("2 days before Easter") \
            .with_flags("NRV")

        self.define_holiday() \
            .with_name("Ostern") \
            .on("Easter") \
            .with_flags("NRV")

        self.define_holiday() \
            .with_name("Ostermontag") \
            .on("1 day after Easter") \
            .with_flags("NRV")

        self.define_holiday() \
            .with_name("Christi Himmelfahrt") \
            .on("39 days after Easter") \
            .with_flags("NRV")

        self.define_holiday() \
            .with_name("Pfingstsonntag") \
            .on("49 days after Easter") \
            .with_flags("NRV")

        self.define_holiday() \
            .with_name("Pfingstmontag") \
            .on("50 days after Easter") \
            .with_flags("NRV")

        self.define_holiday() \
            .with_name("Fronleichnam") \
            .in_regions(["BW", "BY", "HE", "NW", "RP", "SL"]) \
            .on("60 days after Easter") \
            .with_flags("RV")

        """
        11 days before 4. sunday before 12-25: Buß- und Bettag
        """
        self.define_holiday() \
            .with_name("Buß- und Bettag") \
            .in_region("SN") \
            .on(self.buss_und_bettag) \
            .with_flags("RV")

        """
        2020-05-08: 75. Jahrestag der Befreiung vom Nationalsozialismus und der Beendigung des Zweiten Weltkrieges in Europa
        Introduced 2019 for Berlin
        https://gesetze.berlin.de/perma?d=jlr-FeiertGBEV6P1
        """
        self.define_holiday() \
            .with_name("75. Jahrestag der Befreiung vom Nationalsozialismus und der Beendigung des Zweiten Weltkrieges in Europa") \
            .in_region("BE") \
            .in_years([2020]) \
            .on("05-08") \
            .with_flags("F")

        """
        03-08: Frauentag
        Introduced 2019 for Berlin
        https://gesetze.berlin.de/perma?d=jlr-FeiertGBEV6P1
        """
        self.define_holiday() \
            .with_name("Internationaler Frauentag") \
            .in_region("BE") \
            .since(2019) \
            .on("03-08") \
            .with_flags("F")

        """
        before 2018: 10-31: [BB, MV, SN, ST, TH] [RF] Reformationstag
        since 2018:  10-31: [BB, HB, HH, MV, NI, SH, SN, ST, TH] [RF] Reformationstag
        2017:        10-31: [NRF] Reformationstag (national holiday because of 500th anniversary)

        """
        self.define_holiday() \
            .with_name("Reformationstag") \
            .in_years([2017]) \
            .on("10-31") \
            .with_flags("NRF")

        self.define_holiday() \
            .with_name("Reformationstag") \
            .in_regions(["BB", "MV", "SN", "ST", "TH"]) \
            .until(2016) \
            .on("10-31") \
            .with_flags("RF")

        self.define_holiday() \
            .with_name("Reformationstag") \
            .in_regions(["BB", "HB", "HH", "MV", "NI", "SH", "SN", "ST", "TH"]) \
            .since(2018) \
            .on("10-31") \
            .with_flags("RF")

    @staticmethod
    def buss_und_bettag(year):
        return SmartDayArrow(year, 12, 25).shift_to_weekday("sunday", order=4, reverse=True).shift(days=-11)



from dateutil.easter import EASTER_WESTERN

from .holidays import Country


class AT(Country):
    id = "AT"
    languages = ["de"]
    default_lang = "de"
    regions = ["1", "2", "3", "4", "5", "6", "7", "8", "9"]
    easter_type = EASTER_WESTERN

    def __init__(self):
        super().__init__()

        self.define_holiday() \
            .with_name("Neujahr") \
            .on("01-01") \
            .with_flags("NF")

        self.define_holiday() \
            .with_name("Heilige drei Könige") \
            .on("01-06") \
            .with_flags("NRF")

        self.define_holiday() \
            .with_name("Josef") \
            .in_regions(["2", "6", "7", "8"]) \
            .on("03-19") \
            .with_flags("RF")

        self.define_holiday() \
            .with_name("Staatsfeiertag") \
            .on("05-01") \
            .with_flags("NF")

        self.define_holiday() \
            .with_name("Florian") \
            .in_regions(["4"]) \
            .on("05-04") \
            .with_flags("F")

        self.define_holiday() \
            .with_name("Mariä Himmelfahrt") \
            .on("08-15") \
            .with_flags("NRF")

        self.define_holiday() \
            .with_name("Rupert") \
            .in_regions(["5"]) \
            .on("09-24") \
            .with_flags("F")

        self.define_holiday() \
            .with_name("Tag der Volksabstimmung") \
            .in_regions(["2"]) \
            .on("10-10") \
            .with_flags("F")

        self.define_holiday() \
            .with_name("Nationalfeiertag") \
            .on("10-26") \
            .with_flags("NF")

        self.define_holiday() \
            .with_name("Allerheiligen") \
            .on("11-01") \
            .with_flags("NRF")

        self.define_holiday() \
            .with_name("Martin") \
            .in_regions(["1"]) \
            .on("11-11") \
            .with_flags("F")

        self.define_holiday() \
            .with_name("Leopold") \
            .in_regions(["9", "3"]) \
            .on("11-15") \
            .with_flags("F")

        self.define_holiday() \
            .with_name("Mariä Empfängnis") \
            .on("12-08") \
            .with_flags("NRF")

        self.define_holiday() \
            .with_name("Heiliger Abend") \
            .on("12-24") \
            .with_flags("NRF")

        self.define_holiday() \
            .with_name("Christtag") \
            .on("12-25") \
            .with_flags("NRF")

        self.define_holiday() \
            .with_name("Stefanitag") \
            .on("12-26") \
            .with_flags("NF")

        self.define_holiday() \
            .with_name("Silvester") \
            .on("12-31") \
            .with_flags("NF")

        self.define_holiday() \
            .with_name("Karfreitag") \
            .on("2 days before Easter") \
            .with_flags("NRV")

        self.define_holiday() \
            .with_name("Ostersonntag") \
            .on("Easter") \
            .with_flags("NRV")

        self.define_holiday() \
            .with_name("Ostermontag") \
            .on("1 day after Easter") \
            .with_flags("NRV")

        self.define_holiday() \
            .with_name("Christi Himmelfahrt") \
            .on("39 days after Easter") \
            .with_flags("NRV")

        self.define_holiday() \
            .with_name("Pfingstsonntag") \
            .on("49 days after Easter") \
            .with_flags("NRV")

        self.define_holiday() \
            .with_name("Pfingstmontag") \
            .on("50 days after Easter") \
            .with_flags("NRV")

        self.define_holiday() \
            .with_name("Fronleichnam") \
            .on("60 days after Easter") \
            .with_flags("NRV")



from dateutil.easter import EASTER_WESTERN

from holidata.utils import SmartDayArrow
from .holidays import Country


class CO(Country):
    id = "CO"
    languages = ["es"]
    default_lang = "es"
    easter_type = EASTER_WESTERN

    def __init__(self):
        super().__init__()

        self.define_holiday() \
            .with_name("Año Nuevo") \
            .on("01-01") \
            .with_flags("NF")

        self.define_holiday() \
            .with_name("Día del Trabajo") \
            .on("05-01") \
            .with_flags("NF")

        self.define_holiday() \
            .with_name("Grito de Independencia") \
            .on("07-20") \
            .with_flags("NF")

        self.define_holiday() \
            .with_name("Batalla de Boyacá") \
            .on("08-07") \
            .with_flags("NF")

        self.define_holiday() \
            .with_name("Inmaculada Concepción") \
            .on("12-08") \
            .with_flags("NRF")

        self.define_holiday() \
            .with_name("Navidad") \
            .on("12-25") \
            .with_flags("NRF")

        self.define_holiday() \
            .with_name("Jueves Santo") \
            .on("3 days before Easter") \
            .with_flags("NRV")

        self.define_holiday() \
            .with_name("Viernes Santo") \
            .on("2 days before Easter") \
            .with_flags("NRV")

        self.define_holiday() \
            .with_name("Domingo de Pascua") \
            .on("Easter") \
            .with_flags("NRV")

        self.define_holiday() \
            .with_name("La Ascensión del Señor") \
            .on("43 days after Easter") \
            .with_flags("NRV")

        self.define_holiday() \
            .with_name("Corpus Christi") \
            .on("64 days after Easter") \
            .with_flags("NRV")

        self.define_holiday() \
            .with_name("El Sagrado Corazón de Jesús") \
            .on("71 days after Easter") \
            .with_flags("NRV")

        self.define_holiday() \
            .with_name("Día de los Reyes Magos") \
            .on(self.first_monday_after(1, 6)) \
            .with_flags("NRV")

        self.define_holiday() \
            .with_name("Día de San José") \
            .on(self.first_monday_after(3, 19)) \
            .with_flags("NRV")

        self.define_holiday() \
            .with_name("San Pedro y San Pablo") \
            .on(self.first_monday_after(6, 29)) \
            .with_flags("NRV")

        self.define_holiday() \
            .with_name("Asunción de la Virgen") \
            .on(self.first_monday_after(8, 15)) \
            .with_flags("NRV")

        self.define_holiday() \
            .with_name("Día de la Raza") \
            .on(self.first_monday_after(10, 12)) \
            .with_flags("NV")

        self.define_holiday() \
            .with_name("Todos los Santos") \
            .on(self.first_monday_after(11, 1)) \
            .with_flags("NRV")

        self.define_holiday() \
            .with_name("Independencia de Cartagena") \
            .on(self.first_monday_after(11, 11)) \
            .with_flags("NV")

    @staticmethod
    def first_monday_after(month, day):
        def wrapper(year):
            return SmartDayArrow(year, month, day).shift_to_weekday("monday", including=True)

        return wrapper



from dateutil.easter import EASTER_WESTERN

from .holidays import Country


class PT(Country):
    id = "PT"
    languages = ["pt"]
    default_lang = "pt"
    easter_type = EASTER_WESTERN

    def __init__(self):
        super().__init__()

        self.define_holiday() \
            .with_name("Ano Novo") \
            .on("01-01") \
            .with_flags("NF")

        self.define_holiday() \
            .with_name("Dia da Liberdade") \
            .on("04-25") \
            .with_flags("NF")

        self.define_holiday() \
            .with_name("Dia do Trabalhador") \
            .on("05-01") \
            .with_flags("NF")

        self.define_holiday() \
            .with_name("Dia de Portugal") \
            .on("06-10") \
            .with_flags("NF")

        self.define_holiday() \
            .with_name("Assunção de Nossa Senhora") \
            .on("08-15") \
            .with_flags("NF")

        self.define_holiday() \
            .with_name("Implantação da República") \
            .on("10-05") \
            .with_flags("NF")

        self.define_holiday() \
            .with_name("Dia de Todos os Santos") \
            .on("11-01") \
            .with_flags("NF")

        self.define_holiday() \
            .with_name("Restauração da Independência") \
            .on("12-01") \
            .with_flags("NF")

        self.define_holiday() \
            .with_name("Imaculada Conceição") \
            .on("12-08") \
            .with_flags("NF")

        self.define_holiday() \
            .with_name("Natal") \
            .on("12-25") \
            .with_flags("NF")

        self.define_holiday() \
            .with_name("Carnaval") \
            .on("47 days before Easter") \
            .with_flags("NRV")

        self.define_holiday() \
            .with_name("Sexta-feira Santa") \
            .on("2 days before Easter") \
            .with_flags("NRV")

        self.define_holiday() \
            .with_name("Páscoa") \
            .on("Easter") \
            .with_flags("NRV")

        self.define_holiday() \
            .with_name("Corpo de Deus") \
            .on("60 days after Easter") \
            .with_flags("NRV")



from .ES import *
from .holidays import Country

__all__ = [
    "Country",
    "AT",
    "BE",
    "BR",
    "CA",
    "CA",
    "CH",
    "CO",
    "CZ",
    "DE",
    "DK",
    "EE",
    "FI",
    "FI",
    "FR",
    "GB",
    "GR",
    "HR",
    "HU",
    "IS",
    "IT",
    "NL",
    "NO",
    "NZ",
    "PL",
    "PT",
    "RU",
    "SE",
    "SI",
    "SK",
    "TR",
    "US",
    "ZA",
]



from dateutil.easter import EASTER_WESTERN

from holidata.utils import SmartDayArrow
from .holidays import Country


class IS(Country):
    id = "IS"
    languages = ["is"]
    default_lang = "is"
    easter_type = EASTER_WESTERN

    def __init__(self):
        super().__init__()

        self.define_holiday() \
            .with_name("Nýársdagur") \
            .on("01-01") \
            .with_flags("NRF")

        self.define_holiday() \
            .with_name("Verkalýðsdagurinn") \
            .on("05-01") \
            .with_flags("NF")

        self.define_holiday() \
            .with_name("Þjóðhátíðardagurinn") \
            .on("06-17") \
            .with_flags("NF")

        self.define_holiday() \
            .with_name("Jóladagur") \
            .on("12-25") \
            .with_flags("NRF")

        self.define_holiday() \
            .with_name("Annar dagur jóla") \
            .on("12-26") \
            .with_flags("NRF")

        self.define_holiday() \
            .with_name("Skírdagur") \
            .on("3 days before Easter") \
            .with_flags("NRV")

        self.define_holiday() \
            .with_name("Föstudagurinn langi") \
            .on("2 days before Easter") \
            .with_flags("NRV")

        self.define_holiday() \
            .with_name("Páskadagur") \
            .on("Easter") \
            .with_flags("NRV")

        self.define_holiday() \
            .with_name("Annar dagur páska") \
            .on("1 day after Easter") \
            .with_flags("NRV")

        self.define_holiday() \
            .with_name("Uppstigningardagur") \
            .on("39 days after Easter") \
            .with_flags("NRV")

        self.define_holiday() \
            .with_name("Hvítasunnudagur") \
            .on("49 days after Easter") \
            .with_flags("NRV")

        self.define_holiday() \
            .with_name("Annar dagur hvítasunnu") \
            .on("50 days after Easter") \
            .with_flags("NRV")

        self.define_holiday() \
            .with_name("Frídagur verslunarmanna") \
            .on("1. monday in August") \
            .with_flags("NV")

        self.define_holiday() \
            .with_name("Sumardagurinn fyrsti") \
            .on(self.first_thursday_after_04_18) \
            .with_flags("NV")

        """ 
        Both, Christmas Eve (_aðfangadagur jóla_) and New Year's Eve (_gamlársdagur_) are public holidays in Iceland from 13:00 only.
        They're included as full-day holidays, but with an explanatory note.
        """
        self.define_holiday() \
            .with_name("Aðfangadagur jóla") \
            .on("12-24") \
            .with_flags("NRF") \
            .annotated_with("Holiday from 13:00")

        self.define_holiday() \
            .with_name("Gamlársdagur") \
            .on("12-31") \
            .with_flags("NF") \
            .annotated_with("Holiday from 13:00")

    @staticmethod
    def first_thursday_after_04_18(year):
        return SmartDayArrow(year, 4, 18).shift_to_weekday("thursday")



from dateutil.easter import EASTER_WESTERN

from .holidays import Country


class FR(Country):
    id = "FR"
    languages = ["fr"]
    default_lang = "fr"
    easter_type = EASTER_WESTERN

    def __init__(self):
        super().__init__()

        self.define_holiday() \
            .with_name("Jour de l'an") \
            .on("01-01") \
            .with_flags("NF")

        self.define_holiday() \
            .with_name("Fête du premier mai") \
            .on("05-01") \
            .with_flags("NF")

        self.define_holiday() \
            .with_name("Armistice 1945") \
            .on("05-08") \
            .with_flags("NF")

        self.define_holiday() \
            .with_name("Fête nationale") \
            .on("07-14") \
            .with_flags("NF")

        self.define_holiday() \
            .with_name("Assomption") \
            .on("08-15") \
            .with_flags("NRF")

        self.define_holiday() \
            .with_name("Toussaint") \
            .on("11-01") \
            .with_flags("NRF")

        self.define_holiday() \
            .with_name("Armistice 1918") \
            .on("11-11") \
            .with_flags("NF")

        self.define_holiday() \
            .with_name("Noël") \
            .on("12-25") \
            .with_flags("NF")

        self.define_holiday() \
            .with_name("Lundi de Pâques") \
            .on("1 day after Easter") \
            .with_flags("NRV")

        self.define_holiday() \
            .with_name("Ascension") \
            .on("39 days after Easter") \
            .with_flags("NRV")

        self.define_holiday() \
            .with_name("Pentecôte") \
            .on("49 days after Easter") \
            .with_flags("NRV")

        self.define_holiday() \
            .with_name("Lundi de Pentecôte") \
            .on("50 days after Easter") \
            .with_flags("NRV")



import pytest
from snapshottest.file import FileSnapshot
from snapshottest.formatter import Formatter

from holidata import Locale, Country
from tests import HOLIDATA_YEAR_MAX

SNAPSHOT_FILE_PATH_PATTERN = "snapshots/snap_test_holidata/{}[{}-{}] 1.py"


def create_locales():
    response = []

    for country_class in Country.plugins:
        response.extend([Locale(country_class(), lang_id) for lang_id in country_class.languages])

    return response


locales = create_locales()


@pytest.fixture(params=range(2011, HOLIDATA_YEAR_MAX))
def year(request):
    return request.param


@pytest.fixture(params=locales, ids=[loc.id for loc in locales])
def locale(request):
    return request.param


def test_holidata_produces_holidays_for_locale_and_year(snapshot, tmpdir, locale, year):
    temp_file = tmpdir.join(f"{locale.id}.{year}.py")

    export_data = [h.as_dict() for h in locale.get_holidays_of(year)]
    export_data.sort(key=lambda x: (x["date"], x["description"], x["region"]))
    temp_file.write(Formatter().format(export_data, 0))

    try:
        snapshot.assert_match(FileSnapshot(str(temp_file)))
    except AssertionError:
        with open(temp_file, "r") as tf:
            actual = "".join(tf.readlines())

        snapshot_file = SNAPSHOT_FILE_PATH_PATTERN.format(
            "test_holidata_produces_holidays_for_locale_and_year",
            locale.id,
            year)

        with open(snapshot_file) as sf:
            expected = "".join(sf.readlines())

        assert (actual == expected)



from holidata.holidays.holidays import Region


class NC(Region):
    def __init__(self, country):
        super().__init__("NC", country)

        self.define_holiday() \
            .with_name("Lunes siguiente a la Epifanía del Señor") \
            .in_years([2013, 2019]) \
            .on("01-07") \
            .with_flags("RF")

        self.define_holiday() \
            .with_name("San José") \
            .in_years([2012, 2014, 2015, 2019, 2020, 2021]) \
            .on("03-19") \
            .with_flags("RF")

        self.define_holiday() \
            .with_name("Santiago Apóstol / Día Nacional de Galicia") \
            .in_years([2023, 2024]) \
            .on("07-25") \
            .with_flags("RF")

        self.define_holiday() \
            .with_name("Jueves Santo") \
            .on("3 days before Easter") \
            .with_flags("RV")

        self.define_holiday() \
            .with_name("Lunes de Pascua") \
            .on("1 day after Easter") \
            .with_flags("RV")

        self.define_holiday() \
            .with_name("Santiago Apóstol") \
            .in_years([2011, 2013, 2015, 2016, 2017, 2022]) \
            .on("07-25") \
            .with_flags("RF")

        self.define_holiday() \
            .with_name("Lunes siguiente al Día de la Constitución Española") \
            .in_years([2020]) \
            .on("12-07") \
            .with_flags("F")

        self.define_holiday() \
            .with_name("Lunes siguiente a la Natividad del Señor") \
            .in_years([2022]) \
            .on("12-26") \
            .with_flags("RF")

        self.define_holiday() \
            .with_name("San Esteban") \
            .in_years([2011, 2016]) \
            .on("12-26") \
            .with_flags("RF")



from dateutil.easter import EASTER_WESTERN

from .holidays import Country


class NO(Country):
    id = "NO"
    languages = ["nb"]
    default_lang = "nb"
    easter_type = EASTER_WESTERN

    def __init__(self):
        super().__init__()

        self.define_holiday() \
            .with_name("Nyttårsdag") \
            .on("01-01") \
            .with_flags("NF")

        self.define_holiday() \
            .with_name("Offentlig Høytidsdag") \
            .on("05-01") \
            .with_flags("NF")

        self.define_holiday() \
            .with_name("Frigjøringsdag 1945") \
            .on("05-08") \
            .with_flags("NF")

        self.define_holiday() \
            .with_name("Grunnlovsdag") \
            .on("05-17") \
            .with_flags("NF")

        self.define_holiday() \
            .with_name("Julaften") \
            .on("12-24") \
            .with_flags("NRF")

        self.define_holiday() \
            .with_name("Juledag") \
            .on("12-25") \
            .with_flags("NRF")

        self.define_holiday() \
            .with_name("Juledag") \
            .on("12-26") \
            .with_flags("NRF")

        self.define_holiday() \
            .with_name("Nyttårsaften") \
            .on("12-31") \
            .with_flags("NF")

        self.define_holiday() \
            .with_name("Fastelavn") \
            .on("49 days before Easter") \
            .with_flags("NRV")

        self.define_holiday() \
            .with_name("Palmesøndag") \
            .on("7 days before Easter") \
            .with_flags("NRV")

        self.define_holiday() \
            .with_name("Skjærtorsdag") \
            .on("3 days before Easter") \
            .with_flags("NRV")

        self.define_holiday() \
            .with_name("Langfredag") \
            .on("2 days before Easter") \
            .with_flags("NRV")

        self.define_holiday() \
            .with_name("Påskedag") \
            .on("Easter") \
            .with_flags("NRV")

        self.define_holiday() \
            .with_name("Påskedag") \
            .on("1 day after Easter") \
            .with_flags("NRV")

        self.define_holiday() \
            .with_name("Kristi Himmelfartsdag") \
            .on("39 days after Easter") \
            .with_flags("NRV")

        self.define_holiday() \
            .with_name("Pinsedag") \
            .on("49 days after Easter") \
            .with_flags("NRV")

        self.define_holiday() \
            .with_name("Pinsedag") \
            .on("50 days after Easter") \
            .with_flags("NRV")



from dateutil.easter import EASTER_WESTERN

from .holidays import Country


class IT(Country):
    id = "IT"
    languages = ["it"]
    default_lang = "it"
    easter_type = EASTER_WESTERN

    def __init__(self):
        super().__init__()

        self.define_holiday() \
            .with_name("Capodanno") \
            .on("01-01") \
            .with_flags("NF")

        self.define_holiday() \
            .with_name("Epifania") \
            .on("01-06") \
            .with_flags("NRF")

        self.define_holiday() \
            .with_name("Festa della liberazione") \
            .on("04-25") \
            .with_flags("NF")

        self.define_holiday() \
            .with_name("Festa del lavoro") \
            .on("05-01") \
            .with_flags("NF")

        self.define_holiday() \
            .with_name("Festa della repubblica") \
            .on("06-02") \
            .with_flags("NF")

        self.define_holiday() \
            .with_name("Assunzione (ferragosto)") \
            .on("08-15") \
            .with_flags("NRF")

        self.define_holiday() \
            .with_name("Ognissanti") \
            .on("11-01") \
            .with_flags("NRF")

        self.define_holiday() \
            .with_name("Immacolata concezione") \
            .on("12-08") \
            .with_flags("NRF")

        self.define_holiday() \
            .with_name("Natale") \
            .on("12-25") \
            .with_flags("NRF")

        self.define_holiday() \
            .with_name("S.to Stefano") \
            .on("12-26") \
            .with_flags("NRF")

        self.define_holiday() \
            .with_name("Pasqua") \
            .on("Easter") \
            .with_flags("NRV")

        self.define_holiday() \
            .with_name("Pasquetta") \
            .on("1 day after Easter") \
            .with_flags("NRV")



from dateutil.easter import EASTER_WESTERN

from .holidays import Country


class HR(Country):
    id = "HR"
    languages = ["hr"]
    default_lang = "hr"
    easter_type = EASTER_WESTERN

    def __init__(self):
        super().__init__()

        self.define_holiday() \
            .with_name("Nova Godina") \
            .on("01-01") \
            .with_flags("NF")

        self.define_holiday() \
            .with_name("Sveta tri kralja") \
            .on("01-06") \
            .with_flags("NRF")

        self.define_holiday() \
            .with_name("Praznik rada") \
            .on("05-01") \
            .with_flags("NF")

        self.define_holiday() \
            .with_name("Dan državnosti") \
            .on("05-30") \
            .with_flags("NF")

        self.define_holiday() \
            .with_name("Dan antifašističke borbe") \
            .on("06-22") \
            .with_flags("NF")

        self.define_holiday() \
            .with_name("Dan pobjede i domovinske zahvalnosti i Dan hrvatskih branitelja") \
            .on("08-05") \
            .with_flags("NF")

        self.define_holiday() \
            .with_name("Velika Gospa") \
            .on("08-15") \
            .with_flags("NRF")

        self.define_holiday() \
            .with_name("Dan svih svetih") \
            .on("11-01") \
            .with_flags("NRF")

        self.define_holiday() \
            .with_name("Božić") \
            .on("12-25") \
            .with_flags("NRF")

        self.define_holiday() \
            .with_name("Sveti Stjepan") \
            .on("12-26") \
            .with_flags("NRF")

        self.define_holiday() \
            .with_name("Uskrs") \
            .on("Easter") \
            .with_flags("NRV")

        self.define_holiday() \
            .with_name("Uskršnji ponedjeljak") \
            .on("1 day after Easter") \
            .with_flags("NRV")

        self.define_holiday() \
            .with_name("Tijelovo") \
            .on("60 days after Easter") \
            .with_flags("NRV")

        self.define_holiday() \
            .with_name("Dan sjećanja na žrtve Domovinskog rata i Dan sjećanja na žrtvu Vukovara i Škabrnje") \
            .since(2020) \
            .on("11-18") \
            .with_flags("NF")



from dateutil.easter import EASTER_WESTERN

from .holidays import Country


class SI(Country):
    id = "SI"
    languages = ["sl"]
    default_lang = "sl"
    easter_type = EASTER_WESTERN

    def __init__(self):
        super().__init__()

        self.define_holiday() \
            .with_name("Novo leto") \
            .on("01-01") \
            .with_flags("NF")

        """
        From 1955 until May 2012, when the National Assembly of Slovenia passed the Public Finance Balance Act,
        2 January was a work-free day. It was reintroduced in 2017.
        2012<: https://www.uradni-list.si/1/objava.jsp?sop=2012-01-1700
        2016<: https://www.uradni-list.si/1/objava.jsp?sop=2016-01-3568
        """
        self.define_holiday() \
            .with_name("Novo leto") \
            .on("01-02") \
            .except_for([2013, 2014, 2015, 2016]) \
            .with_flags("NF")

        self.define_holiday() \
            .with_name("Prešernov dan") \
            .on("02-08") \
            .with_flags("NF")

        self.define_holiday() \
            .with_name("Dan upora proti okupatorju") \
            .on("04-27") \
            .with_flags("NF")

        self.define_holiday() \
            .with_name("Praznik dela") \
            .on("05-01") \
            .with_flags("NF")

        self.define_holiday() \
            .with_name("Praznik dela") \
            .on("05-02") \
            .with_flags("NF")

        self.define_holiday() \
            .with_name("Dan državnosti") \
            .on("06-25") \
            .with_flags("NF")

        self.define_holiday() \
            .with_name("Marijino vnebovzetje") \
            .on("08-15") \
            .with_flags("NRF")

        self.define_holiday() \
            .with_name("Dan reformacije") \
            .on("10-31") \
            .with_flags("NRF")

        self.define_holiday() \
            .with_name("Dan spomina na mrtve") \
            .on("11-01") \
            .with_flags("NF")

        self.define_holiday() \
            .with_name("Božič") \
            .on("12-25") \
            .with_flags("NF")

        self.define_holiday() \
            .with_name("Dan samostojnosti in enotnosti") \
            .on("12-26") \
            .with_flags("NF")

        self.define_holiday() \
            .with_name("Velikonočna nedelja") \
            .on("Easter") \
            .with_flags("NRV")

        self.define_holiday() \
            .with_name("Velikonočni ponedeljek") \
            .on("1 day after Easter") \
            .with_flags("NRV")

        self.define_holiday() \
            .with_name("Binkošti") \
            .on("50 days after Easter") \
            .with_flags("NRV")



from dateutil.easter import EASTER_WESTERN

from holidata.utils import SmartDayArrow
from .holidays import Country


class SE(Country):
    id = "SE"
    languages = ["sv"]
    default_lang = "sv"
    easter_type = EASTER_WESTERN

    def __init__(self):
        super().__init__()

        self.define_holiday() \
            .with_name("Nyårsdagen") \
            .on("01-01") \
            .with_flags("NF")

        self.define_holiday() \
            .with_name("Trettondedag jul") \
            .on("01-06") \
            .with_flags("NRF")

        self.define_holiday() \
            .with_name("Första maj") \
            .on("05-01") \
            .with_flags("NF")

        self.define_holiday() \
            .with_name("Nationaldagen") \
            .on("06-06") \
            .with_flags("NF")

        self.define_holiday() \
            .with_name("Julafton") \
            .on("12-24") \
            .with_flags("NRF")

        self.define_holiday() \
            .with_name("Juldagen") \
            .on("12-25") \
            .with_flags("NRF")

        self.define_holiday() \
            .with_name("Annandag jul") \
            .on("12-26") \
            .with_flags("NRF")

        self.define_holiday() \
            .with_name("Nyårsafton") \
            .on("12-31") \
            .with_flags("NF")

        self.define_holiday() \
            .with_name("Långfredagen") \
            .on("2 days before Easter") \
            .with_flags("NRV")

        self.define_holiday() \
            .with_name("Påskdagen") \
            .on("Easter") \
            .with_flags("NRV")

        self.define_holiday() \
            .with_name("Annandag påsk") \
            .on("1 day after Easter") \
            .with_flags("NRV")

        self.define_holiday() \
            .with_name("Kristi himmelsfärdsdag") \
            .on("39 days after Easter") \
            .with_flags("NRV")

        self.define_holiday() \
            .with_name("Pingstdagen") \
            .on("49 days after Easter") \
            .with_flags("NRV")

        self.define_holiday() \
            .with_name("Midsommarafton") \
            .on(self.day_before_midsommar) \
            .with_flags("NV")

        self.define_holiday() \
            .with_name("Midsommardagen") \
            .on(self.midsommar) \
            .with_flags("NV")

        self.define_holiday() \
            .with_name("Alla helgons dag") \
            .on(self.saturday_after_10_31) \
            .with_flags("NRV")

    @staticmethod
    def midsommar(year):
        """
        Find the Saturday between 20 and 26 June
        """
        return SmartDayArrow(year, 6, 19).shift_to_weekday("saturday", order=1, reverse=False)

    def day_before_midsommar(self, year):
        return self.midsommar(year).shift(days=-1)

    @staticmethod
    def saturday_after_10_31(year):
        return SmartDayArrow(year, 10, 30).shift_to_weekday("saturday", order=1, reverse=False)



import csv
import io
import json

from holidata.plugin import PluginMount


class Emitter(metaclass=PluginMount):
    type = None

    def __init__(self):
        if self.type is None:
            raise ValueError(f"Emitter {self.__class__.__name__} does not provide its type!")

    @staticmethod
    def get(identifier):
        return Emitter.get_plugin(identifier, "type")

    def output(self, holidays):
        pass


class JsonEmitter(Emitter):
    type = "json"

    def output(self, holidays):
        export_data = [h.as_dict() for h in holidays]
        export_data.sort(key=lambda x: (x["date"], x["description"], x["region"]))
        return "\n".join([json.dumps(h, ensure_ascii=False, sort_keys=False, indent=None, separators=(",", ":")) for h in export_data]) + "\n"


class CsvEmitter(Emitter):
    type = "csv"

    def output(self, holidays):
        export_data = [h.as_dict() for h in holidays]
        export_data.sort(key=lambda x: (x["date"], x["description"], x["region"]))
        result = io.StringIO()

        writer = csv.DictWriter(result,
                                ["locale", "region", "date", "description", "type", "notes"],
                                quoting=csv.QUOTE_ALL,
                                lineterminator="\n")
        writer.writeheader()
        writer.writerows(export_data)

        return result.getvalue()


class YamlEmitter(Emitter):
    type = "yaml"

    def output(self, holidays):
        export_data = [h.as_dict() for h in holidays]
        export_data.sort(key=lambda x: (x["date"], x["description"], x["region"]))

        output = "%YAML 1.1\n"
        output += "---\n"
        for holiday in export_data:
            output += "  holiday:\n"

            for key in ["locale", "region", "date", "description", "type", "notes"]:
                value = holiday[key]

                if value is not None and value != "":
                    output += f"    {key}: {value}\n"
                else:
                    output += f"    {key}:\n"

        output += "...\n"
        return output


class XmlEmitter(Emitter):
    type = "xml"

    def output(self, holidays):
        export_data = [h.as_dict() for h in holidays]
        export_data.sort(key=lambda x: (x["date"], x["description"], x["region"]))

        output = "<?xml version=\"1.0\" encoding=\"UTF-8\" ?>\n"
        output += "<holidays>\n"

        for holiday in export_data:
            output += "  <holiday>\n"

            for key in ["locale", "region", "date", "description", "type", "notes"]:
                value = holiday[key] if key in holiday else ""
                output += f"    <{key}>{value if value is not None else ''}</{key}>\n"

            output += "  </holiday>\n"

        output += "</holidays>\n"
        return output



from dateutil.easter import EASTER_WESTERN

from .holidays import Country


class DK(Country):
    id = "DK"
    languages = ["da"]
    default_lang = "da"
    easter_type = EASTER_WESTERN

    def __init__(self):
        super().__init__()

        self.define_holiday() \
            .with_name("Nytårsdag") \
            .on("01-01") \
            .with_flags("NF")

        self.define_holiday() \
            .with_name("Grundlovsdag") \
            .on("06-05") \
            .with_flags("NF")

        self.define_holiday() \
            .with_name("Juledag") \
            .on("12-25") \
            .with_flags("NRF")

        self.define_holiday() \
            .with_name("Anden juledag") \
            .on("12-26") \
            .with_flags("NRF")

        self.define_holiday() \
            .with_name("Skærtorsdag") \
            .on("3 days before Easter") \
            .with_flags("NRV")

        self.define_holiday() \
            .with_name("Langfredag") \
            .on("2 days before Easter") \
            .with_flags("NRV")

        self.define_holiday() \
            .with_name("Påskedag") \
            .on("Easter") \
            .with_flags("NRV")

        self.define_holiday() \
            .with_name("Anden påskedag") \
            .on("1 day after Easter") \
            .with_flags("NRV")

        self.define_holiday() \
            .with_name("Store bededag") \
            .on("26 days after Easter") \
            .with_flags("NRV")

        self.define_holiday() \
            .with_name("Kristi himmelfartsdag") \
            .on("39 days after Easter") \
            .with_flags("NRV")

        self.define_holiday() \
            .with_name("Pinsedag") \
            .on("49 days after Easter") \
            .with_flags("NRV")

        self.define_holiday() \
            .with_name("Anden pinsedag") \
            .on("50 days after Easter") \
            .with_flags("NRV")



from dateutil.easter import EASTER_WESTERN

from holidata.utils import SmartDayArrow
from .holidays import Country

"""
source: https://almanakka.helsinki.fi/en/flag-days-and-holidays-in-finland.html
"""


class FI(Country):
    id = "FI"
    languages = ["fi", "sv"]
    easter_type = EASTER_WESTERN

    def __init__(self):
        super().__init__()

        self.define_holiday() \
            .with_names({
                "fi": "Uudenvuodenpäivä",
                "sv": "Nyårsdagen",
            }) \
            .on("01-01") \
            .with_flags("NF")

        self.define_holiday() \
            .with_names({
                "fi": "Loppiainen",
                "sv": "Trettondedagen",
            }) \
            .on("01-06") \
            .with_flags("NRF")

        self.define_holiday() \
            .with_names({
                "fi": "Vappu",
                "sv": "Första maj",
            }) \
            .on("05-01") \
            .with_flags("NF")

        self.define_holiday() \
            .with_names({
                "fi": "Itsenäisyyspäivä",
                "sv": "Självständighetsdagen",
            }) \
            .on("12-06") \
            .with_flags("NF")

        self.define_holiday() \
            .with_names({
                "fi": "Joulupäivä",
                "sv": "Juldagen",
            }) \
            .on("12-25") \
            .with_flags("NRF")

        self.define_holiday() \
            .with_names({
                "fi": "Tapaninpäivä",
                "sv": "Annandag jul",
            }) \
            .on("12-26") \
            .with_flags("NRF")

        self.define_holiday() \
            .with_names({
                "fi": "Pitkäperjantai",
                "sv": "Långfredagen",
            }) \
            .on("2 days before Easter") \
            .with_flags("NRV")

        self.define_holiday() \
            .with_names({
                "fi": "Pääsiäispäivä",
                "sv": "Påskdagen",
            }) \
            .on("Easter") \
            .with_flags("NRV")

        self.define_holiday() \
            .with_names({
                "fi": "2. pääsiäispäivä",
                "sv": "Annandag påsk",
            }) \
            .on("1 day after Easter") \
            .with_flags("NRV")

        self.define_holiday() \
            .with_names({
                "fi": "Helatorstai",
                "sv": "Kristi himmelfärdsdag",
            }) \
            .on("39 days after Easter") \
            .with_flags("NRV")

        self.define_holiday() \
            .with_names({
                "fi": "Helluntaipäivä",
                "sv": "Pingst",
            }) \
            .on("49 days after Easter") \
            .with_flags("NRV")

        """
        Saturday between 20 and 26 June: Midsummer Day
        """
        self.define_holiday() \
            .with_names({
                "fi": "Juhannuspäivä",
                "sv": "Midsommardagen",
            }) \
            .on(self.midsummer_day) \
            .with_flags("NRV")

        """
        Saturday between 31 October and 6 November: All Saints' Day
        """
        self.define_holiday() \
            .with_names({
                "fi": "Pyhäinpäivä",
                "sv": "Alla helgons dag",
            }) \
            .on(self.all_saints_day) \
            .with_flags("NRV")

    @staticmethod
    def midsummer_day(year):
        return SmartDayArrow(year, 6, 19).shift_to_weekday("saturday", order=1, reverse=False)

    @staticmethod
    def all_saints_day(year):
        return SmartDayArrow(year, 10, 30).shift_to_weekday("saturday", order=1, reverse=False)



from dateutil.easter import EASTER_WESTERN

from .holidays import Country


class SK(Country):
    id = "SK"
    languages = ["sk"]
    default_lang = "sk"
    easter_type = EASTER_WESTERN

    def __init__(self):
        super().__init__()

        self.define_holiday() \
            .with_name("Deň vzniku Slovenskej republiky") \
            .on("01-01") \
            .with_flags("NF")

        self.define_holiday() \
            .with_name("Zjavenie Pána / Traja králi") \
            .on("01-06") \
            .with_flags("NRF")

        self.define_holiday() \
            .with_name("Sviatok práce") \
            .on("05-01") \
            .with_flags("NF")

        self.define_holiday() \
            .with_name("Deň víťazstva nad fašizmom") \
            .on("05-08") \
            .with_flags("NF")

        self.define_holiday() \
            .with_name("Sviatok svätého Cyrila a Metoda") \
            .on("07-05") \
            .with_flags("NRF")

        self.define_holiday() \
            .with_name("Výročie SNP") \
            .on("08-29") \
            .with_flags("NF")

        self.define_holiday() \
            .with_name("Deň Ústavy Slovenskej republiky") \
            .on("09-01") \
            .with_flags("NF")

        self.define_holiday() \
            .with_name("Sedembolestná Panna Mária") \
            .on("09-15") \
            .with_flags("NRF")

        self.define_holiday() \
            .with_name("Sviatok všetkých svätých") \
            .on("11-01") \
            .with_flags("NRF")

        self.define_holiday() \
            .with_name("Deň boja za slobodu a demokraciu") \
            .on("11-17") \
            .with_flags("NF")

        self.define_holiday() \
            .with_name("Štedrý deň") \
            .on("12-24") \
            .with_flags("NRF")

        self.define_holiday() \
            .with_name("Prvý sviatok vianočný") \
            .on("12-25") \
            .with_flags("NRF")

        self.define_holiday() \
            .with_name("Druhý sviatok vianočný") \
            .on("12-26") \
            .with_flags("NRF")

        self.define_holiday() \
            .with_name("Veľký piatok") \
            .on("2 days before Easter") \
            .with_flags("NRV")

        self.define_holiday() \
            .with_name("Veľkonočný pondelok") \
            .on("1 day after Easter") \
            .with_flags("NRV")



import re

from holidata.plugin import PluginMount
from holidata.utils import SmartDayArrow, month_reference, easter


class Holiday:
    """
    A sheer container for one holiday.
    """

    def __init__(self, locale, region, date, description, flags="", notes=""):
        self.locale = locale
        self.region = region
        self.date = date
        self.description = description
        self.flags = flags
        self.notes = notes

    def as_dict(self):
        return {
            "locale": self.locale,
            "region": self.region,
            "date": self.date.strftime("%Y-%m-%d"),
            "description": self.description,
            "type": self.flags,
            "notes": self.notes
        }


class HolidayGenerator(object):
    name_dict = dict()
    default_lang = None
    country_id = None

    fixed_regex = re.compile(
        r"^\s*(?P<month>\d\d)-(?P<day>\d\d)$",
        re.UNICODE,
    )
    nth_weekday_regex = re.compile(
        r"^\s*(?P<order>\d+)\.(?P<last> last | )(?P<weekday>[a-z]+) in (?P<month>[a-zA-Z]+)$",
        re.UNICODE,
    )
    easter_shift_regex = re.compile(
        r"^\s*((?P<days>\d+) day(s)? (?P<direction>(before|after)) )?Easter$",
        re.UNICODE,
    )

    def __init__(self, country_id, default_lang, easter_type):
        self.name_dict = dict()
        self.notes = ""
        self.regions = [""]
        self.flags = ""
        self.date = None
        self.country_id = country_id
        self.default_lang = default_lang
        self.easter_type = easter_type
        self.filters = []

    def with_name(self, name, lang=None):
        lang = self.default_lang if lang is None else lang
        self.name_dict[lang] = name
        return self

    def with_names(self, name_dict):
        self.name_dict = name_dict
        return self

    def on(self, date):
        self.date = date
        return self

    def with_flags(self, flags):
        self.flags = flags
        return self

    def in_regions(self, regions):
        self.regions = regions
        return self

    def in_region(self, region):
        self.regions = [region]
        return self

    def annotated_with(self, note):
        self.notes = note
        return self

    def in_years(self, reference):
        def reference_does_contain(year):
            return year in reference

        self.filters.append(reference_does_contain)

        return self

    def since(self, reference):
        def reference_is_less_or_equal_to(year):
            return reference <= year

        self.filters.append(reference_is_less_or_equal_to)
        return self

    def until(self, reference):
        def reference_is_greater_or_equal_to(year):
            return reference >= year

        self.filters.append(reference_is_greater_or_equal_to)
        return self

    def except_for(self, reference):
        def reference_does_not_contain(year):
            return year not in reference

        self.filters.append(reference_does_not_contain)

        return self

    def on_condition(self, condition):
        self.filters.append(condition)

        return self

    def _create_date(self, date, year):
        if callable(date):
            return date(year)

        function_map = [
            (self.fixed_regex, self._date_from_fixed_reference),
            (self.nth_weekday_regex, self._date_from_weekday_reference),
            (self.easter_shift_regex, self._date_from_easter_reference),
        ]

        for reg_exp, create_date_from in function_map:
            m = reg_exp.search(date)
            if m is not None:
                return create_date_from(m, year)

        return None

    def build_for_year_and_lang(self, year, lang):
        for filter in self.filters:
            if not filter(year):
                return []

        for region in self.regions:
            yield Holiday(
                locale=f"{lang}-{self.country_id}",
                region=region,
                date=self._create_date(self.date, year),
                description=self.name_dict[lang],
                flags=self.flags,
                notes=self.notes,
            )

    @staticmethod
    def _date_from_fixed_reference(m, year):
        return SmartDayArrow(year, int(m.group("month")), int(m.group("day")))

    @staticmethod
    def _date_from_weekday_reference(m, year):
        return month_reference(
            year, m.group("month"), first=m.group("last").strip() == ""
        ).shift_to_weekday(
            m.group("weekday"),
            order=int(m.group("order")),
            reverse=m.group("last").strip() == "last",
            including=True,
        )

    def _date_from_easter_reference(self, m, year):
        if self.easter_type is None:
            raise ValueError(f"Country '{self.country_id}' does not provide its easter type! Should be one of [WESTERN|ORTHODOX].")

        return easter(year, self.easter_type).shift(
            days=int((m.group("days")) if m.group("days") is not None else 0) * (1 if m.group("direction") == "after" else -1)
        )


class Country(metaclass=PluginMount):
    """
    Represents holidays of a given country
    """
    id = None
    languages = []
    default_lang = None
    easter_type = None
    holiday_generators = []

    def __init__(self):
        self.holiday_generators = []
        self.regions = []

        if self.id is None:
            raise ValueError(f"Country '{self.__class__.__name__}' does not provide an id!")

        if not self.languages:
            raise ValueError(f"Country '{self.__class__.__name__}' does not list any languages!")

        if self.default_lang is not None and self.default_lang not in self.languages:
            raise ValueError(f"Country '{self.__class__.__name__}' does not list language '{self.default_lang}'!")

    @staticmethod
    def get(identifier):
        return Country.get_plugin(identifier, "id")

    def validate_language_or_get_default(self, lang_id):
        if lang_id is not None and lang_id.lower() not in self.languages:
            raise ValueError(
                f"Language '{lang_id}' is not defined for country '{self.id}'! Choose one of [{', '.join(self.languages)}].")
        elif lang_id is None and self.default_lang is not None:
            lang_id = self.default_lang
        elif lang_id is None:
            raise ValueError(
                f"Country '{self.id}' has no default language specified! Choose one of [{', '.join(self.languages)}].")

        return lang_id.lower()

    def define_holiday(self):
        generator = HolidayGenerator(self.id, self.default_lang, self.easter_type)
        self.holiday_generators += [generator]
        return generator

    def get_holidays_of(self, year, lang):
        for generator in self.holiday_generators:
            for holiday in generator.build_for_year_and_lang(year, lang):
                yield holiday

        for region in self.regions:
            for holiday in region.get_holidays_of(year, lang):
                yield holiday


class Region:
    def __init__(self, id, country):
        self.id = id
        self.country = country
        self.holiday_generators = []

    def define_holiday(self):
        generator = HolidayGenerator(self.country.id, self.country.default_lang, self.country.easter_type)
        generator.in_region(self.id)
        self.holiday_generators += [generator]
        return generator

    def get_holidays_of(self, year, lang):
        for generator in self.holiday_generators:
            for holiday in generator.build_for_year_and_lang(year, lang):
                yield holiday
